@@ -2,6 +2,7 @@ from influxdb_client import InfluxDBClient
 from influxdb_client.client.write_api import SYNCHRONOUS
 from influxdb_client import WritePrecision
 from influxdb_client.client.exceptions import InfluxDBError
+from influxdb_lite.attributes import Field, Tag
 import datetime as dt
 import time
 import logging
@@ -14,17 +15,20 @@ class Client(InfluxDBClient):
         self.url = url
         self.token = token
         self.org = org
-        self.query_str = ''
+        self.query_list = []
         self.measurement = None
         self.select_list = ['_time']
 
+    def _query_str(self):
+        return '\n'.join(self.query_list)
+
     def query(self, measurement):
         """Defines the base query from the bucket and the name of the measurement selected. All the following
         methods need a base query to work. """
         self.measurement = measurement
         self.select_list += measurement.tags + measurement.fields
-        self.query_str = '\n'.join([f'from(bucket: "{measurement.bucket}")',
-                                   f'|> filter(fn: (r) => r._measurement == "{measurement.name}")'])
+        self.query_list = [f'from(bucket: "{measurement.bucket}")',
+                           f'|> filter(fn: (r) => r._measurement == "{measurement.name}")']
         return self
 
     def select(self, *args, method: str = 'or'):
@@ -33,20 +37,22 @@ class Client(InfluxDBClient):
         self._check_attr(args, _type='fields')
         arg_names = [arg.name for arg in args]
         self.select_list = arg_names if '_time' in arg_names else arg_names + ['_time']
-        query_list = self.query_str.split('\n')
-        range_idxs = [i for i in range(len(query_list)) if 'range' in query_list[i]]
+        range_idxs = [i for i in range(len(self.query_list)) if 'range' in self.query_list[i]]
         range_idx = 1 if not range_idxs else range_idxs[0]+1
-        query_list.insert(
+        self.query_list.insert(
             range_idx,
             self._contain_or_or(column="_field", _list=self.select_list, method=method)
         )
-        self.query_str = '\n'.join(query_list)
         return self
 
     def _check_attr(self, args, _type: str = 'columns'):
         """Checks is args correspond to either fields, tags or columns"""
         if _type not in ('columns', 'fields', 'tags'):
             raise ValueError(f"Unrecognized type {_type} to check inside of. ")
+        for arg in args:
+            if not isinstance(args, (Field, Tag)):
+                raise AttributeError(f'The attributes passed should be an instance of the classes Tag or Field. Received '
+                                     f'{type(args)} instead.')
         for arg in args:
             if arg.name not in getattr(self.measurement, _type):
                 raise ValueError(f"Field: {arg.name} is not present in measurement {self.measurement}")
@@ -58,10 +64,8 @@ class Client(InfluxDBClient):
         If 'stop' key is not present or its value is None, now() will be considered as default. If 'start' key is not
         present method will raise an error. """
         self._validate_selection(['_time'])
-        query_list = self.query_str.split('\n')
         v_start, v_stop = self._validate_range(start, stop)
-        query_list.insert(1, f"|> range(start: {v_start}, stop: {v_stop})")
-        self.query_str = '\n'.join(query_list)
+        self.query_list.insert(1, f"|> range(start: {v_start}, stop: {v_stop})")
         return self
 
     def _validate_range(self,  start: (int, float, str, dt.datetime), stop: (int, float, str, dt.datetime)):
@@ -79,7 +83,7 @@ class Client(InfluxDBClient):
             raise ValueError(f"Invalid start value. ")
         elif isinstance(start, float):
             start = int(start)
-        elif isinstance(stop, int) or isinstance(stop, str):
+        elif isinstance(start, int) or isinstance(start, str):
             pass
         elif isinstance(start, dt.datetime):
             start = self._dt_to_RFC3339(start)
@@ -94,37 +98,31 @@ class Client(InfluxDBClient):
         * The 'in_' operation for fields must be used in conjunction with the select method and only one field at a time
          to work properly.
         """
-        query_list = self.query_str.split('\n')
         for (attr, comparator, value) in args:
             if attr in self.measurement.tags:
                 if comparator != 'in':
-                    query_list.append(f'|> filter(fn: (r) => r["{attr}"] {comparator} "{value}")')
+                    self.query_list.append(f'|> filter(fn: (r) => r["{attr}"] {comparator} "{value}")')
                 else:
-                    query_list.append(self._contain_or_or(column=attr, _list=value, method=method))
+                    self.query_list.append(self._contain_or_or(column=attr, _list=value, method=method))
             elif attr in self.measurement.fields:
                 if comparator != 'in':
-                    query_list.append(f'|> filter(fn: (r) => r["_field"] == "{attr}" and r["_value"] {comparator} {value})')
+                    self.query_list.append(f'|> filter(fn: (r) => r["_field"] == "{attr}" and r["_value"] {comparator} {value})')
                 else:
-                    query_list.append(self._contain_or_or(column="_value", _list=value, method=method))
+                    self.query_list.append(self._contain_or_or(column="_value", _list=value, method=method))
             else:
                 ValueError(f"Unrecognized attribute {attr} given in dictionary.")
-        self.query_str = '\n'.join(query_list)
         return self
 
     def group_by(self, _list: list):
         """Group by the influxdb tables based on influxdb columns. """
         self._validate_selection(_list)
-        query_list = self.query_str.split('\n')
-        query_list.append(f'|> group(columns: {self._parse_list_into_str(_list)})')
-        self.query_str = '\n'.join(query_list)
+        self.query_list.append(f'|> group(columns: {self._parse_list_into_str(_list)})')
         return self
 
     def order_by(self, _list: list, desc: bool):
         """Sorts influxdb columns in descending or ascending order. """
         self._validate_selection(_list)
-        query_list = self.query_str.split('\n')
-        query_list.append(f'|> sort(columns: {self._parse_list_into_str(_list)}, desc: {str(desc).lower()})')
-        self.query_str = '\n'.join(query_list)
+        self.query_list.append(f'|> sort(columns: {self._parse_list_into_str(_list)}, desc: {str(desc).lower()})')
         return self
 
     def pivot(self, row_keys: list = None, column_keys: list = None, value_column: str = '_value'):
@@ -132,47 +130,39 @@ class Client(InfluxDBClient):
         a sql-like table. """
         row_keys = ['_time'] if row_keys is None else row_keys
         column_keys = ['_field'] if column_keys is None else column_keys
-        query_list = self.query_str.split('\n')
-        query_list.append(f'|> pivot(rowKey:{self._parse_list_into_str(row_keys)}, columnKey: {self._parse_list_into_str(column_keys)}, valueColumn: "{value_column}")')
-        self.query_str = '\n'.join(query_list)
+        self.query_list.append(f'|> pivot(rowKey:{self._parse_list_into_str(row_keys)}, columnKey: {self._parse_list_into_str(column_keys)}, valueColumn: "{value_column}")')
         return self
 
     def limit(self, lmt: int):
         """Limits the amount of results to {lmt}. """
-        query_list = self.query_str.split('\n')
-        query_list.append(f'|> limit(n:{lmt})')
-        self.query_str = '\n'.join(query_list)
+        self.query_list.append(f'|> limit(n:{lmt})')
         return self
 
-    def last(self, *args):
-        """Returns the last non-null records from selected columns. """
-        if len(args) != 1:
-            raise ValueError(f"Only one column can be used for last column. ")
-        self._check_attr(args)
-        query_list = self.all(return_before_execute=True).query_str.split('\n')
-        query_list.append(f'|> last(column:"{args[0].name}")')
-        self.query_str = '\n'.join(query_list)
+    def last(self, field: Field = None):
+        """Applies the last() command from FluxQL. """
+        column = f'column:"{field.name}"' if field else ''
+        self.query_list.append(f'|> last({column})')
+        return self
+
+    def exec(self):
         return self._execute()
 
     def _execute(self):
-        return self._tables_iterator(self.query_api().query(query=self.query_str, org=self.org))
+        return self._tables_iterator(self.query_api().query(query=self._query_str(), org=self.org))
 
-    def all(self, return_before_execute: bool = False):
-        out = self.drop(['_start', '_stop']).pivot()
-        return out if return_before_execute else out._execute()
+    def all(self):
+        return self.drop(['_start', '_stop']).pivot().exec()
 
     def raw(self):
         """Executes the resulting query. """
-        return self.drop(['_start', '_stop'])._execute()
+        return self.drop(['_start', '_stop']).exec()
 
     def drop(self, _list: list):
-        query_list = self.query_str.split('\n')
-        query_list.append(f'|> drop(columns:{self._parse_list_into_str(_list)})')
-        self.query_str = '\n'.join(query_list)
+        self.query_list.append(f'|> drop(columns:{self._parse_list_into_str(_list)})')
         return self
 
     def to_dataframe(self):
-        return self.drop(['_start', '_stop']).query_api().query_data_frame(self.query_str)
+        return self.drop(['_start', '_stop']).query_api().query_data_frame(self._query_str())
 
     def _contain_or_or(self, column: str, _list: list, method: str = 'contains'):
         if method == 'contains':
@@ -234,44 +224,79 @@ class Client(InfluxDBClient):
             return len(isoformat.split('.')[1])
 
     def bulk_insert(self, measurements: list, precision: str = 'ns', write_mode: str = 'SYNCHRONOUS'):
-        """ Receives a list of measurement objects and inserts them at the same time. At least one tag and one
+        """
+        Receives a list of measurement objects or dictionaries and inserts them in bulk. At least one tag and one
         field per measure are needed. Empty-valued tags or fields will not be included.
-        Sets the precision to either seconds (s), milliseconds (ms), microseconds(us) or nanoseconds (default).
-        Precision is set for all the batch inserted. timestamp has to be an int equal to the number of s, ms, us or ns
-        since epoch. For example use time.time_ns() for default precision and int(time.time()) for precision = 's'. """
-        sequence = [''] * len(measurements)
-        bucket = measurements[0].bucket
-        for i in range(len(measurements)):
-            values = measurements[i].get_values()
-            tag_set = ','.join([f"{tag}={values[tag]}"
-                                for tag in measurements[i].tags if values.get(tag, None) is not None])
-            field_set = ','.join([f"{field}={values[field]}"
-                                  for field in measurements[i].fields if values.get(field, None) is not None])
+
+        measurements: Can be either measurement objects or dictionaries. Dictionaries expect 'bucket', 'name',
+                      'fields', 'tags' to be defined. '_time' can be optionally given. Fields and tags also expect
+                      dictionaries as an input.
+        precision:    Sets the precision to either seconds (s), milliseconds (ms), microseconds(us) or nanoseconds
+                      (default). Precision is set for all the batch inserted. timestamp has to be an int equal to the
+                      number of s, ms, us or ns since epoch. For example use time.time_ns() for default precision and
+                      int(time.time()) for precision = 's'.
+        """
+        if isinstance(measurements[0], dict):
+            self._bulk_insert_dicts(measurements, precision=precision, write_mode=write_mode)
+        else:
+            self._bulk_insert_measurements(measurements, precision=precision, write_mode=write_mode)
+
+    def _bulk_insert_dicts(self, measurements: list, precision: str = 'ns', write_mode: str = 'SYNCHRONOUS'):
+        sequence = []
+        bucket = measurements[0]['bucket']
+        for measure in measurements:
+            if bucket != measure['bucket']:
+                raise ValueError('Bulk insert is only supported for one bucket at a time')
+            tag_set = ','.join(f"{tag}={measure['tags'][tag]}" for tag in measure['tags'])
+            field_set = ','.join(f'{field}="{measure["fields"][field]}"'
+                                 if isinstance(measure["fields"][field], str)
+                                 else f'{field}={measure["fields"][field]}' for field in measure['fields'])
             if tag_set == '' or field_set == '':
-                raise ValueError(f"Cannot insert zero fields nor tags in measurement number {i}. ")
-            if values.get('_time', None) is not None:
-                sequence[i] = f"{measurements[i].name},{tag_set} {field_set} {values['_time']}"
+                raise ValueError(f'Cannot insert zero fields in measurement.')
+            if measure.get('_time', None) is not None:
+                sequence.append(f'{measure["name"]},{tag_set} {field_set} {measure["_time"]}')
             else:
-                sequence[i] = f"{measurements[i].name},{tag_set} {field_set}"
+                sequence.append(f'{measure["name"]},{tag_set} {field_set}')
+        self._write_batch(bucket, sequence, precision=precision, write_mode=write_mode)
+
+    def _bulk_insert_measurements(self, measurements: list, precision: str = 'ns', write_mode: str = 'SYNCHRONOUS'):
+        sequence = []
+        bucket = measurements[0].bucket
+        for measure in measurements:
+            if bucket != measure.bucket:
+                raise ValueError('Bulk insert is only supported for one bucket at a time')
+            values = measure.get_values()
+            tag_set = ','.join(f'{tag}={values[tag]}'
+                               for tag in measure.tags if values.get(tag, None) is not None)
+            field_set = ','.join(f'{field}="{values[field]}"'
+                                 if isinstance(values[field], str)
+                                 else f'{field}={values[field]}'
+                                 for field in measure.fields if values.get(field, None) is not None)
+            if tag_set == '' or field_set == '':
+                raise ValueError(f'Cannot insert zero fields in measurement.')
+            if values.get('_time', None) is not None:
+                sequence.append(f'{measure.name},{tag_set} {field_set} {values["_time"]}')
+            else:
+                sequence.append(f'{measure.name},{tag_set} {field_set}')
+        self._write_batch(bucket, sequence, precision=precision, write_mode=write_mode)
+
+    def _write_batch(self, bucket: str, batch: list, precision: str = 'ns', write_mode: str = 'SYNCHRONOUS'):
         if write_mode == 'SYNCHRONOUS':
             with self.write_api(write_options=SYNCHRONOUS) as write_api:
-                write_api.write(bucket=bucket, org=self.org, record='\n'.join(sequence),
+                write_api.write(bucket=bucket, org=self.org, record='\n'.join(batch),
                                 write_precision=getattr(WritePrecision, precision.upper()))
         elif write_mode == 'ASYNCHRONOUS':
             with self.write_api(success_callback=self.on_success, error_callback=self.on_error,
                                 retry_callback=self.on_retry) as write_api:
-                write_api.write(bucket=bucket, org=self.org, record='\n'.join(sequence),
+                write_api.write(bucket=bucket, org=self.org, record='\n'.join(batch),
                                 write_precision=getattr(WritePrecision, precision.upper()))
 
-    def _tables_iterator(self, tables, yield_result: bool = True):
-        """Implements a iterator over resulting tables of a query so that the user can easily iterate the resulting
+    def _tables_iterator(self, tables):
+        """Implements an iterator over resulting tables of a query so that the user can easily iterate the resulting
         rows"""
-        if yield_result:
-            for table in tables:
-                for record in table.records:
-                    yield self.cast_types(record.values)
-        else:
-            return [self.cast_types(record.values) for table in tables for record in table.records]
+        for table in tables:
+            for record in table.records:
+                yield self.cast_types(record.values)
 
     def cast_types(self, values: dict):
         for key in values:
@@ -280,10 +305,10 @@ class Client(InfluxDBClient):
         return values
 
     def on_success(self, conf: (str, str, str), data: str):
-        self.logger.info(f"Written batch: {conf}, data: {data}")
+        pass
 
     def on_error(self, conf: (str, str, str), data: str, exception: InfluxDBError):
         self.logger.error(f"Cannot write batch: {conf}, data: {data} due: {exception}")
 
     def on_retry(self, conf: (str, str, str), data: str, exception: InfluxDBError):
-        self.logger.warning(f"Retryable error occurs for batch: {conf}, data: {data} retry: {exception}")
\ No newline at end of file
+        pass
